/**
 * api机制通讯类
 *
 * 需要econn实例或其仿制实例，具体方法：
 *    sendApiMsg       发送apiMsg消息，需支持直接发送二进制流
 *    log,logDebug  输出日志
 *
 * 需apiFunc方法集，或可给唯一的func
 *
 * 使用范围：前端、后端
 *
 * api调用错误代码约定：
 *
 * 0：   无错误
 * 20：  请求格式错误，未尝试发出调用请求
 * 21：  apiConn实例重置时，请求被抛弃
 * 40：  未能收到反馈，超时timeout
 * 100+： 执行方反馈的错误代码
 *
 */

// 内部引用
const protoWork=require('./protoWork');

class c_apiConn {
  constructor(econn,apiFunc) {
    // 记录发送方法、日志名
    this.econn=econn;
    // 唯一函数处理方法，
    if(typeof apiFunc==='function') this.staticFunc=apiFunc;
    // 函数搜索类方法
    else this.apiFunc=apiFunc;
    /**
     * 是否处在初始态
     *
     * 初始态定义：
     * 1、未接收到apiMsg消息；
     * 2、未成功发出apiMsg消息。
     * 3、一旦发生1或2的情况，初始态结束。
     *
     * 是否初始态区别：
     * 1、仅非初始态响应reset方法，响应后置为初始态。
     */
    this.isInit=true; // 成功发出消息、收到响应，会变为false。为true时，不响应reset动作。

    this.protoWork=protoWork;

    // 对外调用count计数器
    this._count = 1;
    // 己方对外请求的缓存
    this._reqCache = {
      // key 发出时己方生成的count
      // value 数组三项：[sec, apiReqBuf, callback]
    };
    // 对方请求，己方运行完毕的结果缓存
    this._resCache = {
      // key 对方请求的count
      // value 数组三项：[sec, apiResBuf]
      // 其中，json项为null，则表明己方接收到请求但尚未处理完毕；为object时，表明已经处理完毕
    };
    // 是否已被销毁
    this._destroyed=false;
    // 内部秒循环
    this._secLoopClock=setInterval(()=>this._secLoopFunc(),1000);
  }

  // 日志输出相关
  log(str){
    this.econn.log('API - '+str);
  }
  logDebug(str){
    if(this.debug) this.econn.log('API [Debug] - '+str);
  }

  // 秒循环工作函数
  _secLoopFunc() {
    // 如果已经销毁，不再进行任何处理
    if(this._destroyed) return;

    /* 缓存定时清理 */
    // 响应结果保存30秒，超过时间删除
    for (const count in this._resCache) {
      // 获取发送对象
      const resObj = this._resCache[count];
      // 计时器累加
      resObj[0]++;
      // 超时删除
      if (resObj[0] >= 30) delete this._resCache[count];
    }

    // 请求收到反馈时已经删除，未收到时在缓存里
    for (const count in this._reqCache) {
      // 获取发送对象
      const reqObj = this._reqCache[count];
      // 计时器累加
      reqObj[0]++;
      // 第5、15秒重新发送
      if (reqObj[0] === 10 || reqObj[0] === 20) {
        const succ=this.econn.sendApiMsg(reqObj[1]);
        // 凡是成功发出消息，结束初始态
        if(succ) this.isInit=false;
        // 调试输出
        this.logDebug(`REQ 重发请求，${succ?'成功':'失败'}，count: ` + count);
      }
      // 超过30秒删除，执行回调函数
      if (reqObj[0] >= 30) {
        delete this._reqCache[count]; // 从字典中删除
        reqObj[2] && reqObj[2](this._errMaker(30,'timeout')); // 执行回调函数，报超时错误

        // 调试输出
        this.logDebug(`REQ 请求超时，count: ` + count);
      }
    }
  }

  // 生成错误对象
  _errMaker(code,message,list){
    const protoErr=protoWork.lookupType('ProtoBuf.Plant3.model_err');
    // 实例化消息
    const mid=protoErr.create({
      code, message,
      list:list?list.map(i=>i.toString()):[]
    });
    // 反馈
    return mid;
  }

  // 获取请求序号
  _getCount(){
    // 获取序号
    const count = this._count;
    if (count < 214e7) this._count++; // 计数器累加
    else this._count = 1; // 越界归1

    return count;
  }

  // 接到对方调用API请求
  _rec_apiReq(apiReq){
    // 如果已经销毁，不再接收消息
    if(this._destroyed) return;

    // 普通 API调用请求
    /* 根据count判断该此请求要怎么处理 */
    const resCache = this._resCache[apiReq.count];

    // 新请求，缓存字典中无记录，着手处理该请求
    if(!resCache){
      // 调试输出
      this.logDebug(`RES 新请求(${apiReq.api})，建立缓存，执行，count: `+apiReq.count);
      // 首先在resCache字典中，建立记录
      const resCache = this._resCache[apiReq.count]=[0,null,null];

      // 获取处理函数
      const func=this.staticFunc||this.apiFunc.getFunc(apiReq.api);

      // 构造info对象
      const info={
        econn:this.econn,
        api:apiReq.api,
        heads:{}
      };
      // 写入heads信息
      for(let i=0;i<apiReq.heads.length;i+=2) info.heads[apiReq.heads[i]]=apiReq.heads[i+1];

      // API执行callback，或Return返回err信息后的处理，只能执行一次
      let isRun=false;
      const afterCalk=(...all)=>{
        // 多次响应抛出错误，为编程错误
        if(isRun) throw new Error('该API的回调，或Return带有err信息，被执行了多次');
        isRun=true;

        // 如果有err项目，则typ必须是0(undefined)或model_err模板
        if(all.length>0){
          // 错误项判定为false时，一律写成undefined，格式化后typ为0;
          if(!all[0]) all[0]=undefined;
          // 有错误内容，构造model_err模板
          else{
            const err=all[0];
            // 数组类型
            if(Array.isArray(err)) all[0]=this._errMaker(100,undefined,err);
            // Int类型
            else if(Number.isSafeInteger(err)) all[0]=this._errMaker(err||100);
            // 字符串类型
            else if(typeof err==='string') all[0]=this._errMaker(100,err);
            // 字典类型，可能会throw，为编程错误
            else all[0]=this._errMaker(err.code||100,err.message,err.list);
          }
        }

        // 构造反馈体
        const apiRes = {
          count: -apiReq.count, // 负数表示反馈
          params:all.map(b=>this.protoWork.valueEncode(b))
        };

        // 将apiRes整体序列化
        const apiResBuf=this.protoWork.encode(apiRes,'ProtoBuf.Plant3.bufMsg_3_api');

        // 将反馈结果存入缓存
        resCache[1]=apiResBuf;

        // 将处理结果发送出去
        const succ=this.econn.sendApiMsg(apiResBuf);
        // 凡是成功发出消息，结束初始态
        if(succ) this.isInit=false;
        // 调试输出
        this.logDebug(`RES 处理完毕，初次发送，${succ?'成功':'失败'}，count: ` + apiReq.count);
      };

      // 执行，err接收直接retrun返回的错误信息，如果有，则直接触发执行afterCalk
      const err=func(info,apiReq.params.map(b=>this.protoWork.valueDecode(b)),afterCalk);
      if(err) afterCalk(err);
    }
    // 已经接受到请求，且已经处理完毕
    else if (resCache[1]){
      // 再次发送处理结果
      const succ=this.econn.sendApiMsg(resCache[1]);
      // 凡是成功发出消息，结束初始态
      if(succ) this.isInit=false;
      // 调试输出
      this.logDebug(`RES 重复请求，已处理完毕，再次发送，${succ?'成功':'失败'}，count: ` + apiReq.count);
    }
    // 已经接受到请求，但尚未处理完毕，不用任何处理
    else {
      // 调试输出
      this.logDebug('RES 重复请求，尚未处理完毕，不响应，count: '+apiReq.count);
    }
  }

  // 接到己方请求的反馈
  _rec_apiRes(apiRes){
    // 如果已经销毁，不再接收消息
    if(this._destroyed) return false;

    // 从缓存取出请求对象
    const reqObj=this._reqCache[apiRes.count];
    // 取出成功进行相关操作，失败则意味着已经超时，不用管了
    if(reqObj){
      // 从缓存中删除该记录
      delete this._reqCache[apiRes.count];

      // 有回调函数，解析执行
      if(reqObj[2]){
        reqObj[2](...apiRes.params.map(b=>this.protoWork.valueDecode(b)));
      }

      // 调试输出
      this.logDebug('REQ 请求响应成功，count: '+apiRes.count);
    }
  }

  // econn由不可用变可用时，立即重发所有请求
  sendAble(){
    if(this._destroyed) return;
    let isAny=false;

    // 遍历请求缓存，重新发送
    for(const count in this._reqCache){
      if(this.econn.sendApiMsg(this._reqCache[count][1])){
        // 凡是成功发出消息，结束初始态
        this.isInit=false;
      }
      isAny=true;
    }
    // 有发送成功的情况，输出日志
    if(isAny) this.log('econn变可用，重发当前所有缓存流');
  }

  /**
   * 重置方法，仅非初始态时进行实际操作
   *
   * 实际操作内容：
   * 1、请求序号归1；
   * 2、回归初始态；
   * 3、未获得反馈的请求，callback函数报错；
   * 4、清空请求、响应缓存。
   *
   * 应执行reset的场景：
   *  econn的verProc发生变化。
   *  econn从最初无verProc变到有verProc，此时apiConn必定处于初始态，可调用reset，实际无影响。
   */
  reset() {
    // 如果在初始态，不操作
    if(this.isInit) return;

    this.logDebug('reset - 实例重置');

    // 置为初始态
    this.isInit=true;

    // 计数器归1
    this._count=1;

    // 清除对外调用缓存，未执行的callback报错
    for (const count in this._reqCache) {
      const callback = this._reqCache[count][2];
      callback(this._errMaker(21,'apiConn Reset'));
      // 删除
      delete this._reqCache[count];
    }
    // 清除对方前来调用缓存
    for (const count in this._resCache)
      delete this._resCache[count];
  }

  // apiMsg消息入口
  recApiMsg(apiObj){
    // 如果已经销毁，不再进行任何处理
    if(this._destroyed) return;
    // 凡是收到API消息，结束初始态
    this.isInit=false;
    // 请求处理
    if(apiObj.count>0) this._rec_apiReq(apiObj);
    // 反馈处理
    else{
      apiObj.count=-apiObj.count;
      this._rec_apiRes(apiObj);
    }
  }

  /**
   * 调用对方API请求
   *
   * @param apiObj string-API名；object-详细字典
   * @param params 参数数组，若为单个，则内部包装成数组
   * @param callback 回调函数，可以为空
   */
  callApi(apiObj,params,callback) {
    // 如果已经销毁，不再进行任何处理
    if(this._destroyed){
      callback(this._errMaker(20,'apiConn Destroyed'));
      return;
    }

    /* api请求信息解析 */
    let api,heads;
    // 字符串格式的支持
    if(typeof apiObj==='string'){
      heads={};
      api=apiObj;
    }
    // 对象形式的支持
    else {
      api=apiObj.api;
      heads=apiObj.heads||{};
    }

    // 请求参数处理，必须是数组
    if(!Array.isArray(params)) params=[params];

    // 包装请求体
    const reqJson={
      count:this._getCount(),
      api,
      heads:[],
      params:new Array(params.length)
    };

    // 写入请求头信息
    for(const key in heads){
      if(!heads.hasOwnProperty(key)) continue;// 跳过非属性项
      const value=heads[key];
      const typ=typeof value;
      let str;
      // 可传输的类型
      if(typ==='number'||typ==='boolean') str=value.toString();
      else if(typ==='string') str=value;
      else if(value===null) str='null';
      // 不可传输，跳过
      else continue;

      // 写入
      reqJson.heads.push(key,str)
    }

    // 写入参数
    for(let i=0;i<params.length;i++)
      reqJson.params[i]=protoWork.valueEncode(params[i]);
    // 生成流
    const apiReqBuf=this.protoWork.encode(reqJson,'ProtoBuf.Plant3.bufMsg_3_api');
    // 缓存请求
    this._reqCache[reqJson.count] = [0, apiReqBuf,callback];
    // 初次发送请求
    const succ=this.econn.sendApiMsg(apiReqBuf);
    // 凡是成功发出消息，结束初始态
    if(succ) this.isInit=false;
    // 写调试日志
    this.logDebug(`REQ 发出新请求，${succ?'成功':'失败'}，count: `+reqJson.count);
  }

  // 销毁
  destroy() {
    // 标记已经销毁
    this._destroyed=true;
    // 清除定时器
    clearInterval(this._secLoopClock);
    // 调用重置方法
    this.reset();
  }
}

module.exports=c_apiConn;
