/**
 * 服务间通讯类
 *
 * 需netPoint类或其仿制类：
 *    apiFunc 方法集实例，可给唯一函数
 *    id      己方的id
 *
 * 需targetId，对方id
 *
 * 使用范围：前端、后端
 */

// 外部引用
const etools=require('../east-tools');

// 内部引用
const c_apiConn=require('./c_apiConn');

// 进程通讯类
class c_econn {
  constructor(netPoint,targetId){
    /* 常量 */
    this.netPoint= netPoint; // 记录网点对象
    this.apiFunc = netPoint.apiFunc;  // 函数方法集
    this.selfId  = netPoint.id;       // 自己的id
    this.id      = targetId;          // 对方的id
    this.api     = new c_apiConn(this,this.apiFunc); // api机制类
    this.wsUrl     = null; // 前端使用的webSocket连接地址
    // 在后端使用时，需生成connKey
    if(etools.isNode) this.connKey = require('crypto').randomBytes(16);

    // 调用一次重置方法
    this.reset()
  }

  // 重置，构造函数及意识到对方verProc变化时触发
  reset(){
    // 清空verProc记录
    this.verProc=0;
    // api机制类需重置
    this.api.reset();
    // 如果之前有连接池，销毁其中所有socket
    if(this.tcpPool) this.tcpPool.forEach(socket=>socket.destroy('verProc passed'));
    // 重建连接池
    this.tcpPool=[];
    // 重试次数，下次重连时间归零
    this.tryCount=0;
    this.tryAgainSt=0;
  }

  // 日志输出相关
  get _showName(){
    return this.selfId+'.econn: '+this.id;
  }
  log(str){
    etools.log(`[${this._showName}] `+str);
  }
  logDebug(str){
    if(isDebug) etools.log(`[${this._showName} Debug] `+str);
  }

  // 登记新socket
  recIssueSocket(socket){
    // 是否初次启动，是否重建连接
    let isBoot,isUp;

    // 如果verProc发生变化，重置
    if(this.verProc && socket.verProc!==this.verProc) this.reset();
    // 如果verProc为0，则表明是新上线
    if(!this.verProc) isBoot=true;

    // 向连接池追加，以verSock属性排序，以保证两端连接池顺序一致
    etools.orderInsert(this.tcpPool,socket,'verSock');
    // 如果追加后只有一项，说明是从不可用变可用，告知API机制发送缓存
    if(this.tcpPool.length===1){
      isUp=true;
      this.api.sendAble();
    }
    // 记录此次verProc
    this.verProc=socket.verProc;
    // 尝试次数归零
    this.tryCount=0;

    // 接收该socket的onApiMsg方法
    socket.onApiMsg=(apiMsg)=>this.api.recApiMsg(apiMsg);

    // 接收该socket的onclose方法，为和前端ws原生保持一致，全小写
    socket.onclose=()=>{
      // 在当前池中定位该socket
      const index=this.tcpPool.indexOf(socket);
      // 没有定位到，可结束
      if(index<0) return;
      // 从池中移除
      this.tcpPool.splice(index,1);
      // 如果池中为空，尝试调用onDown方法
      if(this.tcpPool.length===0 && this.netPoint.onDown) this.netPoint.onDown(this);
    };

    // onBoot和onUp方法尝试调用
    if(isBoot){
      if(this.netPoint.onBoot) this.netPoint.onBoot(this);
      if(this.netPoint.onUp)   this.netPoint.onUp(this);
    }
    else if(isUp){
      if(this.netPoint.onUp) this.netPoint.onUp(this);
    }
  }

  // 发送apiMsg
  sendApiMsg(apiMsg){
    // 是否发送成功
    let isSendOk=false; // 默认不成功
    // 使用TcpPool中的任一socket发送出去即可
    for(let i=0; this.tcpPool && !isSendOk && i<this.tcpPool.length; i++){
      // 尝试发送并修改标记
      try{
        this.tcpPool[i].sendApiMsg(apiMsg);
        isSendOk=true;
      } catch (e) {}
    }
    // 返回是否成功
    return isSendOk;
  }

  // 是否可用
  get usable(){
    return this.tcpPool.length>0;
  }

  // 调用API方法交给API机制类
  callApi(...all){
    this.api.callApi(...all);
  }
}

module.exports=c_econn;
