/**
 * 事件机制类，依赖于netPoint或其仿制实例
 *
 * 需用到netPoint中的以下方法、对象：
 *  apiFunc.reg   用于注册特定方法函数
 *  callApi       用于向订阅者推送值
 *  log,logDebug  用于输出日志
 *
 * 使用范围：前端、后端
 */
class c_event{
  constructor(netPoint){
    this.debug=true;
    // 事件值获取函数
    this._netPoint=netPoint;

    // 对方订阅请求
    netPoint.apiFunc.reg('_eventBook',(...all)=>this._api_eventBook(...all));
    // 对方触发事件
    netPoint.apiFunc.reg('_eventEmit',(...all)=>this._api_eventEmit(...all));

    // 己方向外订阅字典
    /*
    * key为全事件名，含对方服务名
    * 值字典中：
    *   err
    *   state     0 订阅失败或未订阅，1 订阅中，2 订阅API调用正常
    *   serviceId
    *   eventName
    *   funcs      - 数组，订阅函数集合
    *   paramsArry - 订阅结果数组，既最近emit数组，若为null则表明之前没有emit过
    * */
    this._books={};

    // 外方向己方订阅的事件名字典
    /*
     * key为己方事件名
     * value字典中：
     *    paramsArry：值数组
     *    bookers：订阅服务id数组
     */
    this._emits={};

    /**
     * 订阅验证函数
     *
     * 默认在err位置返回false,，即允许任何订阅
     * @param eventName 事件名
     * @param id 订阅者标识
     * @param callback
     */
    this.bookCheck=(eventName,id,callback)=>{
      callback();
    };

    /**
     * 彻底取消订阅，即该事件没有任何id在订阅
     */
    this.cancel=(eventName)=>{};
  }

  // 订阅请求处理
  _api_eventBook(info,[eventName],callback){
    const id=info.econn.id;
    // 调用验证函数，判定是否允许订阅
    this.bookCheck(eventName,id,(err)=>{
      // 不允许订阅则结束，允许则callback时err实际为空
      callback(err);
      if(err) return;

      // 调试日志
      this._netPoint.log(`[Event Booked] ${eventName} @ ${id}`);

      /* 记录订阅 */
      // emit字典中没有记录，创建
      let emitObj=this._emits[eventName];
      if(!emitObj) emitObj=this._emits[eventName]={bookers:[]};

      // 有数组，且未包含该id，添加
      if(!emitObj.bookers.includes(id)) emitObj.bookers.push(id);

      // 如果当前推送记录有值，进行一次推送，不考虑在此推送失败的情况
      if(emitObj.paramsArry) info.econn.callApi('_eventEmit',[eventName,...emitObj.paramsArry],()=>{});
    })
  }

  // 对方触发事件
  _api_eventEmit(info,[eventName,...paramsArry],callback){
    const event=info.econn.id+'.'+eventName;
    const bookObj=this._books[event];
    // 如果己方没有订阅，则callback错误
    if(!bookObj) callback('noBook');
    // 己方有订阅
    else {
      // 记录本次推送的结果
      bookObj.paramsArry=paramsArry;
      // 记录最初数组长度
      const len=bookObj.funcs.length;
      // 严格按照最初的数组长度，确定循环次数
      for(let i=0;i<len;i++){
        // 从头部取出一个func
        const func=bookObj.funcs.shift();
        // 执行，若反馈判定为真，再添加回尾部，判定为假，不再加入，即已删除
        if(func(...paramsArry)) bookObj.funcs.push(func);
      }
      // 如果最终数组长度大于0，空反馈，即继续订阅
      if(bookObj.funcs.length>0) callback();
      // 没有func了，反馈err取消订阅
      else {
        // 取消订阅
        callback('noBook');
        // 删除己方对该事件的记录节点
        delete this._books[event];
      }
    }
  }

  /**
   * 重新定于指定service下的所有事件
   * @param serviceId 服务id
   * @param forceReBook 是否强制重新订阅
   * @private
   *
   */
  reBookService (serviceId,forceReBook) {
    // 遍历所有当前订阅的事件
    for(const id in this._books){
      const bookObj=this._books[id];
      // 判定是否重新订阅
      if(bookObj.serviceId===serviceId && // 服务名与此次通知匹配
        bookObj.state!==1 &&              // 不在订阅过程中
        (forceReBook||bookObj.state===0)  // 强制重新订阅，或当前未订阅
      )
        this._bookCall(bookObj);
    }
  }

  /**
   * 向对应服务发送订阅请求，并更新内部标记状态
   * @param bookObj 订阅状态对象
   * @private
   */
  _bookCall(bookObj){
    // 标记订阅中
    bookObj.state=1;

    // 调用api机制向对方订阅
    this._netPoint.callApi(bookObj.serviceId+'._eventBook',bookObj.eventName,(err)=>{
      // 订阅错误标记
      if(err){
        this._netPoint.logDebug(`[Event - Book Fail   ]: ${bookObj.serviceId}.${bookObj.eventName}`);
        bookObj.state=0;
        bookObj.err=err;
      }
      // 订阅成功
      else {
        this._netPoint.logDebug(`[Event - Book Success]: ${bookObj.serviceId}.${bookObj.eventName}`);
        bookObj.state=2;
        bookObj.err=null;
      }
    })
  }

  // 对外方法：订阅
  book(event,func){
    // 事件名非字符串，程序编写错误
    if((typeof event!=='string')||!event) throw new Error('eventName must be string and not empty');

    // 订阅函数类型
    if(typeof func!=="function") throw new Error('func typeError');

    // 尝试读取已有订阅信息
    let bookObj=this._books[event];

    // 已经订阅，无论订阅状态
    if(bookObj){
      // 缓存func
      bookObj.funcs.push(func);
      // 如果之前曾推送过结果，对本func触发一次
      if(bookObj.paramsArry) func(...bookObj.paramsArry);
    }
    // 新订阅
    else {
      // 分离服务名称和事件名称
      const index=event.indexOf('.');
      if(!index>0) throw new Error('fullEventName: serviceId + eventName');
      const serviceId=event.substr(0,index);
      const eventName=event.substr(index+1);

      // 新建事件本地存储，paramsArry为空
      bookObj=this._books[event]={
        serviceId,eventName,
        state:0,        // 0 未订阅或订阅失败;  1 订阅成功
        verProc:null,     // 最近尝试时的对方版本 null表示未曾订阅，0表示订阅时未知对方时间戳
        err:null,       // 最近一次的错误信息，重新订阅成功时不清除
        funcs:[func],   // 订阅的func数组
        paramsArry:null // 最后一次推送的订阅信息
      };
      // 立即订阅
      this._bookCall(bookObj);
    }
  }

  // 触发某事件的推送，值需要从外部字典获取，若无值，则不必继续
  emit(eventName,...paramsArry){

    /* 记录订阅 */
    // emit字典中没有记录，创建
    let emitObj=this._emits[eventName];
    if(!emitObj) emitObj=this._emits[eventName]={bookers:[]};

    // 记录值，为以后新订阅初次发送准备
    emitObj.paramsArry=paramsArry.length>0?paramsArry:null;

    // 遍历bookers数组进行推送
    emitObj.bookers.forEach(id=>{
      // 调用对方API
      this._netPoint.callApi(id+'._eventEmit',[eventName,...paramsArry],(err)=>{
        // 推送无错误，可结束
        if(!err) return;
        // 推送有错误，判定错误原因

        // 取消订阅
        const index=emitObj.bookers.indexOf(id);
        if(index>=0){
          // 从数组中移除id
          emitObj.bookers.splice(index,1);
          // 调试日志
          this._netPoint.logDebug(`[Event Unbook]: ${eventName} @ ${id}`);
        }
      })
    });
  }
}

module.exports=c_event;
