// 二进制流片段缓存，应对分包粘包问题
/*
* 实例化时，需跟int参数限定最大允许消息体长度
*
* 接收：调用方法bufPieceIn传入
*
* 输出：
*   重新定义bufMsgFunc(bufMsg)方法，接收整理好的标准bufMsg
*   重新定义errFunc(errcode)方法，errcode可能列表如下：
*     overMaxLen    - 超出允许最大长度
*
* 触发errFunc或bufMsgFunc后，均会实例均会复位，可继续使用
* */
class c_fluentCache {
  constructor(maxLen){
    // 默认允许最大64MB
    maxLen=maxLen||67108863;
    // 如果最大长度设定有误，报错
    if(!Number.isSafeInteger(maxLen)||maxLen<1||maxLen>67108863) throw new Error('maxLen Error');
    // 最大允许长度
    this.maxLen=maxLen;
    // 头4位缓存，始终复用
    this.head=Buffer.allocUnsafe(4);
    // 对外接口函数，需在外部替换
    this.errFunc=(errcode)=>{}; // 错误触发函数
    this.bufMsgFunc=(bufMsg)=>{}; // 切开的消息体

    // 初次调用复位方法
    this.reSet();
  }
  // 复位
  reSet(){
    // 当前接收指针，p<4时，长度位未完整；p>4时，接收到body
    this.p=0;
    // 消息体积缓存，长度位接收完毕时创建
    this.msgBuf=null
  }
  // 二进制流片段进入
  bufPieceIn(bufPiece){
    // 验证输入，不是Buffer对象则报错
    if(!Buffer.isBuffer(bufPiece)) throw new Error('bufPiece is not Buffer');

    /* ========== 将二进制片段分析赋值到类内部缓存 ========= */
    for(let i=0;i<bufPiece.length;i++) {
      // 当长度位读取完毕时，构建缓存
      if(this.p===4) {
        // 消息体长度，前4字节的低26位为长度
        const bodyLen=this.head.readUInt32LE() & 67108863; // 相当于二进制的26个1
        // 超过允许最大长度，抛出错误
        if(bodyLen>this.maxLen){
          this.reSet(); // 复位
          this.errFunc('overMaxLen'); // 抛出错误
          break; // 退出循环
        }
        // 未超过，准备接收消息体
        else {
          // 构建消息缓存
          this.msgBuf=Buffer.allocUnsafe(8+bodyLen); // 前置码一共8位，前4位是长度位
          // 将长度位写入
          for(let i=0;i<4;i++) this.msgBuf[i]=this.head[i];
        }
      }


      // p<4时，读取长度位
      if(this.p<4) {
        this.head[this.p]=bufPiece[i];
        this.p++; // 指针位后移
      }
      // p>=4时，读取消息体
      else {
        this.msgBuf[this.p]=bufPiece[i];
        this.p++; // 指针位后移

        // 如果已经接收到最后，触发消息，并复位
        if(this.p===this.msgBuf.length) {
          this.bufMsgFunc(this.msgBuf); // 触发
          this.reSet(); // 复位
        }
      }
    }
  }
}

module.exports=c_fluentCache;