/**
 * 微服务网点类
 *
 *
 *
 * 使用范围：仅后端使用
 */

// 外部引用
const fs=require('fs');
const net=require('net');
const crypto=require('crypto');
const etools=require('../east-tools');

// 内部引用
const isDebug       = require('./isDebug');
const protoWork     = require('./protoWork');
const c_fluentCache = require('./c_fluentCache');
const c_econn       = require('./c_econn');
const c_apiFunc     = require('./c_apiFunc');
const c_event       = require('./c_event');

// 配置
const conf={
  reConnSetpAddMs:2000,
  reConnMaxCount:60
};

/* ====== 明文bufMsg消息处理 ====== */
// 构建bufMsg
function bufMsgBuild(buffer,typeCode){
  /* 输入校验 */
  // buffer类别
  if(!Buffer.isBuffer(buffer)) throw new Error('buffer is not Buffer');
  if(buffer.length>67108863) throw new Error('buffer too long, more then 64MB');
  // code必须是0～64之间
  if(!Number.isSafeInteger(typeCode)||typeCode<0||typeCode>63) throw new Error('code Error');
  /* 构建消息头 */
  const head=Buffer.allocUnsafe(8);

  // 0~3  低26位长度，高6位typeCode
  const b0_4=buffer.length+(typeCode<<26);
  head.writeUInt32LE(b0_4,0);
  // 4    随机位
  head[4]=Math.floor(Math.random()*256);
  // 5~7 异或校验码
  const checkNum=getCheckNum(buffer,b0_4+head[4]);
  head[5]=checkNum%256;
  head[6]=(checkNum>>>8)%256;
  head[7]=checkNum>>>16;

  // 合并head和buffer，即为输出的bufMsg
  return Buffer.concat([head,buffer]);
}
// 异或校验码计算，3字节
function getCheckNum(buffer,seed) {
  seed=seed%16000000;// 不超过2^24即可
  // 遍历buffer计算
  for(let i=0;i<buffer.length;i+=3){
    // 原始数字
    let num=((buffer[i]||0)<<16)+((buffer[i+1]||0)<<8)+(buffer[i+2]||0);
    // 循环位数
    const move=i%8;
    const high=num>>>move;
    const low=num-(high<<move);
    num=high+(low<<(24-move));

    seed=seed^num;
  }
  // 返回
  return seed;
}

const protoApi=protoWork.lookupType('ProtoBuf.Plant3.bufMsg_3_api');
const protoReg=protoWork.lookupType('ProtoBuf.Plant3.bufMsg_4_reg');
class enetFuncs{
  /* === 内部方法 === */
  /**
   * 解析bufMsg
   * @param socket socket对象
   * @param bufMsg 收到的bufMsg流
   * @param level  消息解包级别
   * @returns {*}
   * @private
   *
   * 解析错误时，反馈undefined，socket调用销毁方法
   * 解析成功时，反馈[typecode, msg]
   */
  static _bufMsgDecode(socket,bufMsg,level){
    // 校验码比对
    const b0_4=bufMsg.readUInt32LE(0);
    const buf=bufMsg.slice(8);
    const jsNum=getCheckNum(buf,b0_4+bufMsg[4]);// 计算校验码
    const csNum=bufMsg[5]+(bufMsg[6]<<8)+(bufMsg[7]<<16);// 传输校验码
    // 若比对错误，销毁socket
    if(jsNum!==csNum){
      socket.destroy('checkNum Error');
      return
    }
    // 分离typeCode，并分类处理
    const typeCode=b0_4>>>26;
    switch(typeCode){
      // 明文：API机制，必须已经完成注册
      case 3:
        // 必须已经认证，否则报错
        if(!socket.id)
          socket.destroy('rec apiMsg at not reged');
        // 如果是加密状态的0级，收到明文消息报错
        else if(level===0&&socket.cryptoBuf)
          socket.destroy('rec unknow crypto Msg');
        // 尝试解析protoBuf，成功则触发apiMsg事件
        else {
          // 如果反序列化出错，关闭socket
          try{
            return [3,protoApi.decode(buf)];
          }catch (e) {
            socket.destroy('bufMsg protoApi.decode Error');
          }
        }
        break;
      // 明文：注册消息，仅能在0级未注册状态
      case 4:
        // 必须是0级，未注册
        if(level!==0||socket.id)
          socket.destroy('rec regMsg at Error State');
        // 尝试解析
        else{
          // 如果反序列化出错，关闭socket
          try{
            return [4,protoReg.decode(buf)];
          }catch (e) {
            socket.destroy('bufMsg protoApi.decode Error');
          }
        }
        break;
      // 加密包，解密后迭代
      case 24:
        // 如果socket没有设定秘钥，报错关闭
        if(!socket.cryptoBuf)
          socket.destroy('need crypto seted');
        // 有证书，可以进行解密
        else {
          let newbuf;
          try{
            const decipher = crypto.createDecipher('aes-128-ecb', socket.cryptoBuf);
            const buf_update=decipher.update(buf);
            const buf_final =decipher.final();
            newbuf=Buffer.concat([buf_update,buf_final]);
          } catch (e) {
            socket.destroy('crypto msg decode Error');
          }
          // 解密完成，如果有，迭代调用
          if(newbuf) return this._bufMsgDecode(socket,newbuf,level+1);
        }
        break;
      // 未知类别的包，关闭socket
      default:
        socket.destroy('unknow bufMsg type: '+typeCode);
    }
  }
  // 发送apiMsg、regMsg方法
  static _sendApiMsg(socket,msg){
    // apiMsg支持直接是流，或者是json，在此序列化
    let buf;
    // msg是流，可直接用
    if(Buffer.isBuffer(msg)) buf=msg;
    // 当做普通json尝试
    else {
      // 实例化消息
      const mid=protoApi.create(msg);
      // 取得流并反馈
      buf=protoApi.encode(mid).finish();
    }
    // 构造bufMsg
    const bufMsg=bufMsgBuild(buf,3);

    // 不需要加密
    if(!socket.cryptoBuf){
      // 直接发送明文bufMsg流
      socket.write(bufMsg);
    }
    // 需要加密
    else {
      // 将明文bufMsg流包装一层
      let cryptBuf;
      try{
        let cip = crypto.createCipher('aes-128-ecb', socket.cryptoBuf);
        cryptBuf = Buffer.concat([cip.update(bufMsg),cip.final()]);
      } catch (e){
        // 遇到错误，销毁链接
        socket.destroy('sendApiMsg crypto Error');
        return;
      }

      // 再次包装成bufMsg格式
      const bufMsg2=bufMsgBuild(cryptBuf,24);

      // 发送
      socket.write(bufMsg2);
    }
  }
  static _sendRegMsg(socket,regMsg){
    // regMsg仅支持json发送，在此序列化。socket必须处于未注册状态
    if(socket.id){
      socket.destroy('socket send regMsg at Reged');
      return;
    }

    // 实例化消息
    const mid=protoReg.create(regMsg);
    // 取得流并反馈
    const buf=protoReg.encode(mid).finish();
    // 构造bufMsg
    const bufMsg=bufMsgBuild(buf,4);
    // 发送
    socket.write(bufMsg);
  }
  /* === 对外方法 === */
  // heads节点解析，包装
  static headsDecode(headArry){
    const heads={};
    headArry.forEach(i=>{
      const [key,value]=i.split(':');
      heads[key]=value;
    });
    return heads;
  }
  static headsEncode(heads){
    const arry=[];
    for(const key in heads){
      if(!heads.hasOwnProperty(key)) continue;
      arry.push(key+':'+heads[key]);
    }
    return arry;
  }

  /**
   * 原始socket功能升级
   *
   * 1、添加sendApiMsg方法，必须按照bufMsg_3_api模板格式，支持json对象、bufMsg_3_api实例、或直接buf流
   * 2、未注册时，收到消息触发onRegMsg方法；
   * 3、已注册时，收到消息触发onApiMsg方法；
   * 4、关闭时，触发onclose方法，注意全小写，为了和前端websocket保持一致；
   * 4、执行此函数后，5s后仍处于未注册状态，则销毁；
   * 5、添加empSec属性，每次收到消息归零，但此处不进行递增计算；
   * 6、添加pingpong方法，在外部认为有必要触发心跳时调用。
   */
  static socketUpdate(socket){
    // 分配分包粘包类,默认限制极大允许64兆
    const fluentCache=new c_fluentCache();

    // TCP层数据接收，empSec归零
    socket.empSec=0;// 初始为0
    socket.on('data',(buf)=>{
      socket.empSec=0;// 收到消息归0
      fluentCache.bufPieceIn(buf)
    });
    socket.on('close',()=>{
      // 为兼容Web前端，尝试调用onclose方法
      if(socket.onclose) socket.onclose();
    });

    // 分包粘包输出bufMsg，添加regMsg、apiMsg事件
    fluentCache.bufMsgFunc=(bufMsg)=>{
      // 获取解析bufMsg的结果数组，若为空，则结束
      const arry=enetFuncs._bufMsgDecode(socket,bufMsg,0);
      if(!arry) return;
      // 分离消息格式及消息体
      const [typeCode,msg]=arry;
      // api消息，模板bufMsg_3_api
      if(typeCode===3){
        // 截获特殊命令
        switch (msg.cmd){
          // 心跳包
          case 'ping':socket.sendApiMsg({cmd:'pong'});return;
          case 'pong':return;
        }
        // 非特殊命令，触发onApiMsg方法
        socket.onApiMsg(msg);
      }
      // 注册消息，模板bufMsg_4_reg
      else if(typeCode===4){
        // 触发onRegMsg方法
        socket.onRegMsg(msg);
      }
    };

    // 发送regMsg、apiMsg方法
    socket.sendRegMsg=(msg)=>this._sendRegMsg(socket,msg);
    socket.sendApiMsg=(msg)=>this._sendApiMsg(socket,msg);

    // 发送心跳方法，由外部调用
    socket.pingpong=()=>socket.sendApiMsg({cmd:'ping'});

    // error事件
    socket.on('error',(msg)=>{
      // 关闭日志
      if(isDebug&&(socket.id||msg))
        etools.log(`[socket off] id:${socket.id||''} verProc:${socket.verProc||''} msg:`+msg);
    });

    // 5秒内必须完成注册定时器
    setTimeout(()=>{
      if(!socket.id) socket.destroy('no reg in 5 second');
    },5000);
  }
  // socket添加认证、加密信息
  static socketAuth(socket,id,verProc,skey,cert){
    socket.id=id;
    // verProc标示对方进程版本
    socket.verProc=verProc;
    // skey前8字节转出16位字符串，作为socket版本
    socket.verSock=skey.toString('hex').substr(0,16);
    // 两项都有才进行加密
    if(skey && cert) socket.cryptoBuf=Buffer.concat([cert,skey]);
  }
}

class c_netPoint{
  constructor(){
    // 与其他服务连接进程类
    this._econnDict={};
    // tcp监听实例
    this._tcpServer=null;
    // api方法集合，及内部方法订阅
    this.apiFunc=new c_apiFunc();
    this.apiFunc.reg(/^_{2}/,(info,params,callback)=>{
      const func=this['_api'+info.api];
      // 如果匹配到函数，执行
      if(typeof func==='function') func.apply(this,[info,params,callback]);
      // 未匹配到，反馈错误
      else callback('unknow inner Api');
    });
    // event事件类
    this.event=new c_event(this);
    // 版本时间戳
    this.verProc=Date.now();
    // 启动秒循环函数
    setInterval(()=>this._secLoop(),1000);
  }

  // econn心跳维持方法，在master和service的secLoop中调用
  _socketPingPong(){
    // 遍历所有econn，含master与socket
    for(const id in this._econnDict){
      if(!this._econnDict.hasOwnProperty(id)) continue;
      const econn=this._econnDict[id];
      // 取出tcp连接池，没有则跳过
      const tcpPool=econn.tcpPool;
      if(!tcpPool) continue;
      // 复制连接池，因为在destroy过程中，会对实际连接池进行操作
      const tcpPoolCp=tcpPool.map(i=>{return i});
      // 遍历该连接池中所有socket
      tcpPoolCp.forEach(socket=>{
        // 空闲秒数递增
        socket.empSec++;
        // 超过50秒断线
        if(socket.empSec>50){
          // 调用销毁方法
          socket.destroy();
          // 如果在连接池，则移除
          const index=tcpPool.indexOf(socket);
          if(index>=0) tcpPool.splice(index,1);
        }
        // 超过40秒发心跳
        else if(socket.empSec>40) socket.pingpong();
      })
    }
  }

  // 日志输出相关
  get _showName(){
    return this.id;
  }
  log(str){
    etools.log(`[${this._showName}] `+str);
  }
  logDebug(str){
    if(isDebug) etools.log(`[${this._showName} Debug] `+str);
  }

  // 获取监听端口
  get _tcpPort(){
    if(!this._tcpServer) return 0;
    else return this._tcpServer.address().port;
  }

  // 构建监听端口
  _tcpListenBuild(){
    let tryPort;

    // 获取本次尝试的端口
    const getPort=()=>{
      // 设定为数组
      if(Array.isArray(this._listenPort)){
        // 第一次请求时，取范围下限
        if(!tryPort) tryPort=this._listenPort[0];

        // 如果tryPort未越界，反馈其值并++运算，越界则自然return空
        if(tryPort<=this._listenPort[1]) return tryPort++;
      }
      // 设定为确切数字
      else if(this._listenPort){
        // 如果之前没尝试过，反馈设定值，尝试过，则自然return空
        if(!tryPort) {
          tryPort=this._listenPort;
          return tryPort;
        }
      }
      // 没有设定，使用3w以上的随机值
      else return Math.floor(Math.random()*35536)+30000;
    };

    // 监听成功后的方法
    const listenOk=(server)=>{
      // 登记
      this._tcpServer=server;

      // 日志输出
      this.log(`TCP Listing at: `+this._tcpPort);

      // 新连接到达，调用socket初始化方法
      server.on('connection',(socket)=>this._socketRecOriginal(socket));

      // 如果不是master，则调用连接master方法
      if(this.id!=='master') this._keepMaster();
    };

    // 尝试建立监听方法
    const tryListen=()=>{
      const port=getPort();
      if(!port) throw new Error('build Listen Error');

      const server=net.createServer();    // 创建实例
      server.listen(port);    // 开启监听

      // 是否已判定失败
      let failed=false;

      // 监听关闭、错误事件，均触发再次尝试
      server.on('close',()=>{
        if(!failed){
          failed=true;
          tryListen();
        }
      });
      server.on('error',()=>{
        if(!failed){
          failed=true;
          tryListen();
        }
      });
      // 监听成功，则调用listenOk方法
      server.on('listening',()=>listenOk(server));
    };

    // 初次尝试建立
    tryListen();
  }

  // 经过认证的socket下发到econn实例
  _issueSocket(socket){
    this.log(`issueSocket id:${socket.id} verProc:${socket.verProc} verSock:${socket.verSock} crypto:${!!socket.cryptoBuf}`);

    // 取出当前econn，没有则新建
    let econn=this._econnDict[socket.id];
    if(!econn) econn=this._econnDict[socket.id]=new c_econn(this,socket.id);

    // 登记socket
    econn.recIssueSocket(socket);

    // 如果自己是master，且econn登记完后tcp长度为1，即从不可用变可用，或verProc变化，需发送广播
    if(this.id==='master' && econn.tcpPool.length===1){
      this.broadMsg('__serviceUsable',socket.id,socket.verProc);
    }
  }

  // 调用API方法
  callApi(apiObj,params,callback){
    // 灵活调用格式
    if(typeof params==="function"){
      callback=params;
      params=[];
    }
    if(!callback) callback=()=>{};

    // 字符串封装为标准对象
    if(typeof apiObj==='string') apiObj={api:apiObj};
    // 分离服务名
    const index=apiObj.api.indexOf('.');
    const id=apiObj.api.substr(0,index);
    apiObj.api=apiObj.api.substr(index+1);

    // 获取eConn对象
    const econn=this._geteConn(id);
    // 获取失败，反馈错误（仅master的_geteConn方法，可能获取不到）
    if(!econn) callback('no econn');
    // 获取成功，将api请求转移到对应econn类
    else econn.callApi(apiObj,params,callback);
  }

  // 当前连接的econn字典
  econnDict(){
    const resultDict={};
    for(const id in this._econnDict){
      const econn=this._econnDict[id];
      if(econn && econn.usable) resultDict[id]=econn;
    }
    return resultDict;
  }

  // 判定指定id是否在线
  isIdOn(id){
    const econn=this._econnDict[id];
    if(econn&&econn.usable) return true;
  }
}
class c_master extends c_netPoint{
  constructor(conf){
    super();
    this.id='master';
    this._listenPort=conf.port; // 服务端口，默认随机

    // 是否禁止服务间通讯
    this._forbidMateConn=conf.forbidMateConn;

    // 默认验证函数不加密
    this.regFunc=(post,callback)=>callback(undefined,{});
    // 建立tcp监听，conf中可配置port和host项目
    this._tcpListenBuild();
  }

  // 秒循环函数
  _secLoop(){
    this._socketPingPong()
  }

  // 原始socket功能升级、注册
  _socketRecOriginal(socket){
    // 包装功能
    enetFuncs.socketUpdate(socket);

    // 接收对方发出的注册信息，对方一定是service
    socket.onRegMsg=(regObj)=>{

      // 调用认证函数
      this.regFunc(regObj,(err,json,cert)=>{
        // 有错误，认证失败
        if(err) {
          socket.destroy(err);
          return;
        }

        // 构造反馈消息对象，模板bufMsg_4_req
        const regBack={
          verProc:this.verProc
        };

        // 有证书，需加密
        if(cert){
          regBack.isCrypto=true;
          // 需下发证书
          if(json && json.sendCert){
            regBack.cert=cert;
          }
        }

        // 发送注册反馈消息
        socket.sendRegMsg(regBack);

        // 身份登记
        enetFuncs.socketAuth(socket,regObj.id,regObj.verProc.toNumber(),regObj.skey,cert);

        // 下发至econn实例
        this._issueSocket(socket);

        // 更新注册信息
        this._recordServiceInfo(regObj);
      })
    }
  }

  // 记录或更新服务信息
  _recordServiceInfo(regObj){
    // 获取econn对象，取不到结束
    const econn=this._econnDict[regObj.id];
    if(!econn) return;

    // 整理记录
    econn.regInfo=regObj;
  }

  // master获取econn实例
  _geteConn(id){
    // 唯一来源是可信dict字典，没有则反馈空
    return this._econnDict[id];
  }

  // 服务A想要连接服务B
  _api__connMate(info,[post],callback){
    this.logDebug(`${info.econn.id} 想要连接 ${post.id}`);

    // 如果禁止服务间通讯，此API反馈错误
    if(this._forbidMateConn) {
      callback('master forbidMateConn');
      return;
    }

    // 获取conn字典中，双方的信息
    const conn_a=this._econnDict[info.econn.id];
    const conn_b=this._econnDict[post.id];

    // 如果b不在线，则callback
    if(!conn_b||!conn_b.usable) {
      callback('service off line: '+post.id);
      return;
    }

    // 先要求b尝试连接a
    conn_b.callApi('__tryToMate',{
      _protoBufType:'ProtoBuf.Plant3.bufMsg_4_reg',
      id  : conn_a.id,
      connKey : post.connKey,
      port    : conn_a.regInfo.port,
      innAddr : conn_a.regInfo.innAddr,
      pubAddr : conn_a.regInfo.pubAddr
    },(err,buildObj)=>{
      // 如果有错误发生，反馈
      if(err) callback(err);
      // 如果没有buildObj，表明b连接a成功，此处callback空
      else if(!buildObj) callback();
      // 有buildObj，表明b连a失败，callback时带连接对象，要求a连接b
      else callback(undefined,{
        _protoBufType:'ProtoBuf.Plant3.bufMsg_4_reg',
        id  : conn_b.id,
        connKey : buildObj.connKey,
        port    : conn_b.regInfo.port,
        innAddr : conn_b.regInfo.innAddr,
        pubAddr : conn_b.regInfo.pubAddr
      })
    })
  }

  // 发起广播消息
  broadMsg(broadObj,...params){
    // 整理
    if(typeof broadObj==='string') broadObj={broad:broadObj};
    // 如果没有broad项目，为编程错误
    if(!broadObj.broad||(typeof broadObj.broad!=='string')) throw new Error('broadMsg Format Error');
    // 遍历所有在线可用的socket，调用其api告知
    for(const id in this._econnDict){
      if(!this._econnDict.hasOwnProperty(id)) continue;
      const econn=this._econnDict[id];
      // 当前不可用则不发送
      if(!econn.usable) continue;
      // 可用才发送
      econn.callApi({
        api:'__broadMsg',
        heads:broadObj
      },params,(...all)=>{

      })
    }
  }
}
class c_service extends c_netPoint{
  constructor(conf){
    super();
    this.id          = conf.id;
    this._masterPort = conf.port;
    this._masterHost = conf.host||'127.0.0.1';
    this._certPath   = conf.certPath;
    this._listenPort = conf.listenPort; // 服务端口，默认随机
    this._pubAddr    = conf.pubAddr;    // 公网地址，可以为空

    // 广播处理函数集合，及内部广播接收方法
    this._broadFunc  = new c_apiFunc();
    this._broadFunc.reg(/^_{2}/,(heads,...params)=>{
      const func=this['_broad'+heads.broad];
      // 如果匹配到函数，执行
      if(typeof func==='function') func.apply(this,[heads,...params]);
    });
    // 开启tcp监听
    this._tcpListenBuild();

    // 连接建立与保持
    this._keepMaster();
  }
  // 秒循环函数
  _secLoop(){
    this._keepMaster();
    this._keepMates();
    this._socketPingPong();
  }

  // 创建tcp链接，仅负责创建、连接，连接成功即可callback
  _socketTcpTry(conf,buildObj,callback){
    const socket=net.connect(conf);
    // 建立超时，允许5s
    let timeOut=setTimeout(()=>{
      if(callback) callback('timeout');
      callback=null;
      if(socket) socket.destroy();
    },5e3);
    // 建立成功
    socket.on('connect',()=>{
      clearTimeout(timeOut);
      socket.removeAllListeners(socket.eventNames());
      if(callback){
        this._socketRecOriginal(socket,buildObj);
        callback();
      }
      else socket.destroy();
      callback=null;
    });
    // 建立失败
    socket.on('error',(err)=>{
      clearTimeout(timeOut);
      socket.removeAllListeners(socket.eventNames());
      if(callback) callback(err);
      callback=null;
    });
  }

  // 原始socket功能升级、注册
  _socketRecOriginal(socket,buildObj){
    // 包装功能
    enetFuncs.socketUpdate(socket);

    // 己方发起，发注册消息，等反馈
    if(buildObj){
      // 如果buildObj.id为master时，其他项目均为null
      const targetId=buildObj.id;

      // 发起方设定skey
      const skey=crypto.randomBytes(32);

      // 获取注册消息体
      const regMsg=targetId==='master'?
        this._regMsg_master(socket,skey):
        this._regMsg_mate(socket,skey,buildObj);
      // 发送注册消息
      socket.sendRegMsg(regMsg);
      // 接收注册反馈
      socket.onRegMsg=(regBack)=>{
        // 提取信息
        const verProc=regBack.verProc.toNumber();
        const isCrypto=regBack.isCrypto;
        let   cert=regBack.cert.length>0?regBack.cert:null;

        // 如果target是master，且下发了新证书，则持久化保存
        if(targetId==='master' && cert && this.certPath){
          fs.writeFile(this.certPath,cert,(err)=>{
            if(err) etools.log('通讯证书写磁盘持久化失败，下次连接很可能会失败')
          });
        }

        // 如果要求加密，但没有证书，读取本地证书
        if(isCrypto && !cert){
          try{
            cert = fs.readFileSync(this._certPath);
          } catch (e) {
            socket.destroy('socket crypto Error: no local Cert');
            return;
          }
        }

        // 进行身份登记，此时有无cert可代表是否加密通讯
        enetFuncs.socketAuth(socket,targetId,verProc,skey,cert);

        // 分发给econn实例
        this._issueSocket(socket);
      }
    }
    // 对方发起，等对方注册
    else {
      // 接收对方发出的注册信息，此处对方一定是service
      socket.onRegMsg=(regMsg)=>{
        // 获取econn，无论是对方回连，还是己方回连失败后对方主动来连接，此时都应有econn
        const econn=this._econnDict[regMsg.id];
        if(!econn){
          socket.destroy('no econn Object');
          return;
        }

        // connKey判定
        if(!regMsg.connKey.equals(econn.connKey)){
          socket.destroy('connKey check Error');
          return;
        }

        /* socket验证完毕 */
        // 构造基础反馈对象
        const regBack={
          verProc:this.verProc
        };

        // 发送注册反馈消息
        socket.sendRegMsg(regBack);

        // 进行身份登记，最后的null是证书项
        enetFuncs.socketAuth(socket,regMsg.id,regMsg.verProc.toNumber(),regMsg.skey,null);

        // 分发给econn实例
        this._issueSocket(socket);
      }
    }
  }

  // 获取向对方服务注册的消息体
  _regMsg_master(socket,skey){
    /* 整理注册信息，模板：bufMsg_4_req */
    const regMsg={
      // id标识，计算机名，进程版本
      id:this.id,
      hostname:etools.hostname,
      verProc:this.verProc,
      // 端口记录、内网地址、公网地址
      port:this._tcpPort,
      innAddr:socket.localAddress,
      pubAddr:this._pubAddr,
      // 会话秘钥
      skey
    };

    // 尝试读取本地证书并进行md5运算
    if(this._certPath){
      try{
        const localCert = fs.readFileSync(this._certPath);
        regMsg.certMd5 = etools.md5(localCert);
      } catch (e) {
        this.log('读取本地证书或计算MD5失败，按照没有证书处理');
      }
    }

    return regMsg;
  }
  _regMsg_mate(socket,skey,build){
    /* 整理注册信息，模板：bufMsg_4_req */
    const regMsg={
      // id标识，计算机名，进程版本
      id:this.id,
      hostname:etools.hostname,
      verProc:this.verProc,
      // 会话证书、配对秘钥
      skey,
      connKey:build.connKey
    };

    return regMsg;
  }

  // 保持与master、services的连接，每秒周期调用
  _keepMaster(){
    /* 是否需要进行连接判定 */
    // 无tcpPort时，不执行
    const tcpPort=this._tcpPort;
    if(!tcpPort) return;

    // 取出当前econn，没有则新建
    let econn=this._econnDict.master;
    if(!econn) econn=this._econnDict.master=new c_econn(this,'master');

    // 如果连接池通道数大于0，不必继续
    if(econn.tcpPool.length>0) return;

    /* 进行重新连接操作 */
    const now=Date.now();

    // 当前不允许尝试连接：下次允许时间戳数值，且在可能的设定范围内
    if(econn.tryAgainSt && econn.tryAgainSt>now && econn.tryAgainSt<now+conf.reConnMaxCount*conf.reConnSetpAddMs) return;

    // 计算下次允许尝试的时间戳
    const nextSt=econn.tryAgainSt=now+Math.min(econn.tryCount++,conf.reConnMaxCount)*conf.reConnSetpAddMs-300;

    // 建立TCP连接
    const port=this._masterPort;
    const host=this._masterHost;
    this._socketTcpTry({port,host},{id:'master'},(err)=>{
      // 设定下次允许尝试连接时间
      econn.tryAgainSt=nextSt;
      // 若有错误，输出日志
      if(err) this.log('Conn Master Error: Build TCP Error, count: '+econn.tryCount);
    })
  }
  _keepMates(){
    // 获取master的连接，获取不到或不可用，无法建立
    const econnMaster=this._econnDict.master;
    if(!econnMaster||!econnMaster.usable) return;

    // 遍历所有econn
    const now=Date.now();
    for(const id in this._econnDict){
      // 获取econn，判定跳过条件
      const econn=this._econnDict[id];
      if(
        !econn.id||                             // 跳过无id项目
        econn.usable||                          // 跳过当前正常可用的
        econn.id==='master'||                   // 跳过master，不在此处理
        econn.tryCount>5||                      // 跳过尝试5次依然连不上的服务
        econn.tryAgainSt&&econn.tryAgainSt>now  // 跳过未达到重连时间
      ) continue;

      /* 符合条件，进行重连尝试 */
      econn.tryAgainSt=now+5e3; // 设定下次允许尝试的时间戳
      econn.tryCount++;         // 尝试次数递增

      // 调用master的API协助
      const logStr='请求master协助连接service: '+econn.id+' ';
      this.logDebug(logStr);
      econnMaster.callApi('__connMate',{
        _protoBufType:'ProtoBuf.Plant3.bufMsg_4_reg',
        id:econn.id,       // 要连接对方的id
        connKey:econn.connKey  // 己方econn的connKey，由master传给对方，回连时验证
      },(err,mateBuild)=>{
        // 有错误信息，建立连接失败
        if(err) {
          this.logDebug(logStr+' 失败，err: '+err.message);
        }
        // 有mateBuild，表明需要己方尝试连接对方
        else if(mateBuild) {
          // 尝试连接对方
          this.logDebug(logStr+'需己方主动连接对方');
          this._tryToMate(mateBuild,(err)=>{
            if(err) this.logDebug(logStr+'连接对方失败：'+err.message);
          });
        }
        // 无mateBuild，表明对方尝试连接己方
        else {
          this.logDebug(logStr+'等待对方连接己方');
        }
      });
    }
  }

  /**
   * 主动建立与其他service连接
   *
   * @param mateBuild 对方信息
   * @param callback 回调函数
   */
  _tryToMate(mateBuild,callback){
    // 获取econn，没有则新建
    let econn=this._econnDict[mateBuild.id];
    if(!econn) econn=this._econnDict[mateBuild.id]=
      new c_econn(this,mateBuild.id);

    // 如果已经可用，不再尝试，且反馈成功
    if(econn.usable){
      callback();
      return;
    }

    // 先尝试连接对方内网ip
    this._socketTcpTry({
      port:mateBuild.port,
      host:mateBuild.innAddr
    }, mateBuild, (err)=>{
      // 如果成功，可callback
      if(!err) callback(); // 响应API成功
      // 失败，且对方没有公网接口，反馈失败
      else if(!mateBuild.pubAddr) callback('build Tcp Error: inner fail and no pubAddr');
      // 失败，但对方有公网接口，继续尝试
      else this._socketTcpTry({
          port:mateBuild.port,
          host:mateBuild.pubAddr
        }, mateBuild, (err)=>{
          // 如果成功，可callback
          if(!err) callback(); // 响应API成功
          // 失败，且对方没有公网接口，反馈失败
          else callback('build Tcp Error: inner and pubAddr both Error');
        })
    })
  }

  // 服务获取econn实例
  _geteConn(id){
    // 获取连接实例
    let econn=this._econnDict[id];
    // 获取不到则新建
    if(!econn) econn=this._econnDict[id]=new c_econn(this,id);
    // 反馈
    return econn;
  }

  // 接收master下发广播消息
  _api__broadMsg(info,params,callback){
    // 只允许master调用
    if(info.econn.id!=='master'){
      callback('this API only call by Master');
      return
    }
    // 头信息必须有broad项目
    const broad=info.heads.broad;
    if(!broad) {
      callback('broadMsg API Heads need broad');
      return
    }

    /* 可以发送广播 */
    callback();// 响应成功

    // 获取广播方法
    const func=this._broadFunc.getFunc(broad);
    // 执行
    func(info.heads,...params);
  }

  // master下发要求连接到一般服务
  _api__tryToMate(info,[mateBuild],callback){
    // 只允许master调用
    if(info.econn.id!=='master'){
      callback('this API only call by Master');
      return
    }

    // 调用内部方法，转移callback
    const logStr='应邀主动尝试，连接service: '+mateBuild.id+' ';
    this.logDebug(logStr);
    this._tryToMate(mateBuild,(err)=>{
      if(err){
        // 连接对方失败
        this.logDebug(logStr+'失败，期望对方回连：'+err);
        // 获取econn，此处应有
        const econn=this._econnDict[mateBuild.id];
        // 获取不到，报错
        if(!econn) callback('no econn info');
        // 获取到，构造buildObj，期望对方连接己方
        else callback(undefined,{
          _protoBufType:'ProtoBuf.Plant3.bufMsg_4_reg',
          connKey:econn.connKey
        });
      }
      else callback();
    });
  }

  // 收到服务可用广播，即master认为该服务重新上线
  _broad__serviceUsable(heads,serviceId,verProc){
    // 未曾有过相关信息，什么也不做，此处可排除自己
    const econn=this._econnDict[serviceId];
    if(!econn) return;

    // 该服务允许重连尝试
    econn.tryAgainSt=0;
    econn.tryCount=0;

    // 如果当前有版本记录，且与通知版本不同，则重置
    if(econn.verProc && verProc!==econn.verProc) econn.reset();

    // 如果当前处于不可用状态，重新订阅，刚重置后一定处于不可用
    if(!econn.usable) this.event.reBookService(serviceId,true);
  }
}

module.exports={
  enetFuncs,
  c_master,
  c_service
};